import { Activity, AppApiEvent, AppApiResponse } from './types/appapi.dto.types.ts'
import { FetchesJasdEvents } from '../../usecases/dependencies/FetchesJasdEvents.ts'
import { Logs } from '../../usecases/dependencies/Logs.ts'
import { FetchesJasdActivity } from '../../usecases/dependencies/FetchesJasdActivity.ts'
import { logger } from '../../logging.ts'
import NotFound = Deno.errors.NotFound

export class JasdGateway implements FetchesJasdEvents, FetchesJasdActivity {
  private readonly APP_API_BASE_URL_V1 = 'https://gemeinschaftswerk-nachhaltigkeit.de/app/api/v1'
  private readonly APP_API_BASE_URL_V2 = 'https://gemeinschaftswerk-nachhaltigkeit.de/app/api/v2'
  private readonly APP_FILES_BASE_URL = 'https://gemeinschaftswerk-nachhaltigkeit.de/app/api/v1/files'

  constructor(private readonly logger: Logs) {
  }

  private buildQueryUrl(
    limit: number,
    offset: number,
    localOnly: boolean,
    location?: string,
  ): URL {
    const url = new URL(`${this.APP_API_BASE_URL_V2}/search`)
    url.searchParams.append('page', offset.toString())
    url.searchParams.append('size', limit.toString())
    url.searchParams.append('sort', 'period.start,asc,ignorecase')
    url.searchParams.append('includeExpiredActivities', 'false')
    url.searchParams.append('permanent', 'false')

    if (localOnly) {
      url.searchParams.append('online', 'false')
    }

    if (location) {
      url.searchParams.append('location', location)
    }

    return url
  }

  async fetchActivity(id: string): Promise<Activity> {
    const url = new URL(`${this.APP_API_BASE_URL_V1}/activities/${id}`)

    logger.debug(`Fetching event from ${url}`)
    const response = await fetch(url)

    try {
      const json = await response.json() as Activity
      if (!json) {
        throw new NotFound('Not found')
      }
      return json

      // deno-lint-ignore no-explicit-any
    } catch (e: any) {
      logger.error(
        `Error fetching event ${url}: ${e.message}`,
      )
      throw e
    }
  }

  async fetchEvents(
    limit: number,
    offset: number,
    localOnly: boolean,
    location?: string,
  ): Promise<{ events: AppApiEvent[]; totalCount: number }> {
    const url = this.buildQueryUrl(limit, offset, localOnly, location)
    this.logger.debug(`Fetching events from ${url}`)

    try {
      const response = await fetch(url)
      const data = await response.json() as AppApiResponse

      const events = data.content
        .flat()
        .filter((event) => !event.activity.period.permanent)
        .filter((event) => !!event)

      return {
        events,
        totalCount: data.totalElements,
      }
    } catch (error) {
      this.logger.error(`Error fetching events from ${url}`, error)
      throw error
    }
  }
}
