import { logger } from './logging.ts'
import { GeoApiClient } from './adapters/geo/GeoApiClient.ts'
import { createSchema, createYoga } from './deps.ts'
import { QueryEvents, QueryEventsInput, QueryEventsOutput } from './usecases/QueryEvents.ts'
import { JasdGateway } from './adapters/jasd/JasdGateway.ts'
import { QueryEvent, QueryEventInput, QueryEventOutput } from './usecases/QueryEvent.ts'

const SCHEMA = `
    type Location {
      street: String
      streetNo: String
      zipCode: String
      city: String
    }
    
    type EventDetails {
      description: String!
      externalLink: String
      meetingLink: String
    }
  
    type Event {
      id: String!
      title: String!
      organisationName: String!
      startDate: String
      endDate: String
      location: Location
      isRemote: Boolean!
      imageUrl: String!
      eventDetails: EventDetails!
    }
    
    type EventsResponse {
      totalResults: Int!
      data: [Event]
    }

    type EventResponse {
      data: Event
    }
    
    type Query {
      # uses offset-based pagination as described in https://www.apollographql.com/docs/react/pagination/offset-based
      events(offset: Int = 0, limit: Int = 5, geolocationId: ID, localOnly: Boolean): EventsResponse!
      event(id: String!): EventResponse!
    }
`

export type EventsRequest = {
  limit?: number
  offset?: number
  geolocationId?: string
  localOnly?: boolean
}

export type EventRequest = {
  id: string
}

const validateQueryEventsInput = (request: EventsRequest): QueryEventsInput => {
  logger.debug('validating request: ' + JSON.stringify(request))

  let limit = request.limit ?? 5
  if (limit > 50) {
    limit = 50
  }

  return {
    limit: limit,
    offset: request.offset ?? 0,
    geolocationId: request.geolocationId,
    localOnly: request.localOnly ?? false,
  }
}

const validateQueryEventInput = (request: EventRequest): QueryEventInput => {
  logger.debug('validating request: ' + JSON.stringify(request))
  return {
    id: request.id,
  }
}

const createResolvers = (config: ServerConfig, useCases: UseCases) => ({
  Query: {
    event: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
      parameters: EventRequest,
      // next line is required for the resolver to work
      // deno-lint-ignore no-unused-vars
      context: GraphQLContext,
    ): Promise<QueryEventOutput> => {
      if (config.fake) {
        return Promise.resolve({ data: {} })
      } else {
        const input = useCases.QueryEvent.validateInput(parameters)
        return useCases.QueryEvent.useCase.execute(input)
      }
    },
    events: (
      // deno-lint-ignore no-explicit-any
      _parent: any,
      parameters: EventsRequest,
      // next line is required for the resolver to work
      // deno-lint-ignore no-unused-vars
      context: GraphQLContext,
    ): Promise<QueryEventsOutput> => {
      if (config.fake) {
        return Promise.resolve({ totalResults: 0, data: [] })
      } else {
        const input = useCases.QueryEvents.validateInput(parameters)
        return useCases.QueryEvents.useCase.execute(input)
      }
    },
  },
})

type UseCaseNames = 'QueryEvents' | 'QueryEvent'
type UseCases = Record<UseCaseNames, { useCase: UseCase; validateInput: ValidateInputFn }>
// deno-lint-ignore no-explicit-any
type UseCase = { execute: (input: any) => any }
// deno-lint-ignore no-explicit-any
type ValidateInputFn = (params: any) => any

export const startServer = (config: ServerConfig): Deno.HttpServer<Deno.NetAddr> => {
  const geoApi = new GeoApiClient(config.geoAPIEndpointUrl)
  const jasdEventsGateway = new JasdGateway(logger)

  const useCases = {
    QueryEvent: {
      validateInput: validateQueryEventInput,
      useCase: new QueryEvent(jasdEventsGateway, logger),
    },
    QueryEvents: {
      validateInput: validateQueryEventsInput,
      useCase: new QueryEvents(jasdEventsGateway, geoApi, logger),
    },
  }

  const resolvers = createResolvers(config, useCases)

  const gqlServer = createYoga({
    schema: createSchema({ resolvers, typeDefs: SCHEMA }),
    graphiql: true,
  })

  return Deno.serve({
    port: config.port,
    handler: (req: Request) => {
      const url = new URL(req.url)
      const pathname = url.pathname
      if (pathname.startsWith('/health') || pathname.startsWith('/ready')) return isAlive()
      return gqlServer.handle(req)
    },
    onListen({ port, hostname }) {
      logger.info(
        `Server started at http://${hostname === '0.0.0.0' ? 'localhost' : hostname}:${port}/graphql`,
      )
      if (config.fake) {
        logger.info(
          'Server is serving fake data due to FAKE env var set to true',
        )
      }
    },
  })
}

export const DEFAULT_PORT = 8006

export type ServerConfig = {
  port: number
  fake: boolean // For local development. If set, the API returns dummy data
  imageProxyBaseUrl: string
  geoAPIEndpointUrl: string
}

const isAlive = () => new Response(`${true}`)

type GraphQLContext = {
  request?: { headers: Headers }
  params?: {
    extensions?: {
      headers?: {
        [key: string]: string
      }
    }
  }
}
